from contextlib import contextmanager


def maybe_show_progress(it, show_progress, **kwargs):
    """Optionally show a progress bar for the given iterator.

    Parameters
    ----------
    it : iterable
        The underlying iterator.
    show_progress : bool
        Should progress be shown.
    **kwargs
        Forwarded to the click progress bar.

    Returns
    -------
    itercontext : context manager
        A context manager whose enter is the actual iterator to use.

    Examples
    --------
    .. code-block:: python

       with maybe_show_progress([1, 2, 3], True) as ns:
            for n in ns:
                ...
    """
    if show_progress:
        try:
            import click
        except ImportError as e:
            raise ImportError(
                "click must be installed to show a progressbar"
            ) from e
        return click.progressbar(it, **kwargs)

    @contextmanager
    def ctx():
        yield it

    return ctx()



from functools import lru_cache
from hashlib import md5
import os
import pathlib
import sqlite3
import sys
import logging

import requests

from .beatmap import Beatmap
from .cli import maybe_show_progress


if sys.platform.startswith('win'):
    def sanitize_filename(name):
        for invalid_character in r':*?"\/|<>':
            name = name.replace(invalid_character, '')
        return name
else:
    def sanitize_filename(name):
        return name.replace('/', '')

sanitize_filename.__doc__ = """\
Sanitize a filename so that it is safe to write to the filesystem.

Parameters
----------
name : str
    The name of the file without the directory.

Returns
-------
sanitized_name : str
    The name with invalid characters stripped out.
"""


class Library:
    """A library of beatmaps backed by a local directory.

    Parameters
    ----------
    path : path-like
        The path to a local library directory.
    cache : int, optional
        The amount of beatmaps to cache in memory. This uses
        :func:`functools.lru_cache`, and if set to None will cache everything.
    download_url : str, optional
        The default location to download beatmaps from.
    """
    DEFAULT_DOWNLOAD_URL = 'https://osu.ppy.sh/osu'
    DEFAULT_CACHE_SIZE = 2048

    def __init__(self,
                 path,
                 *,
                 cache=DEFAULT_CACHE_SIZE,
                 download_url=DEFAULT_DOWNLOAD_URL):
        self.path = path = pathlib.Path(path)

        self._cache_size = cache
        self._read_beatmap = lru_cache(cache)(self._raw_read_beatmap)
        self._db = db = sqlite3.connect(str(path / '.slider.db'))
        with db:
            db.execute(
                """\
                CREATE TABLE IF NOT EXISTS beatmaps (
                    md5 BLOB PRIMARY KEY,
                    id INT,
                    path TEXT UNIQUE NOT NULL
                )
                """,
            )
        self._download_url = download_url

    def copy(self):
        """Create a copy suitable for use in a new thread.

        Returns
        -------
        Library
            The new copy.
        """
        return type(self)(
            self.path,
            cache=self._cache_size,
            download_url=self._download_url,
        )

    def close(self):
        """Close any resources used by this library.
        """
        self._read_beatmap.cache_clear()
        self._db.close()

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            # if an error is raised in the constructor
            pass

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    @staticmethod
    def _osu_files(path, recurse):
        """An iterator of ``.osu`` filepaths in a directory.

        Parameters
        ----------
        path : path-like
            The directory to search in.
        recurse : bool
            Recursively search ``path``?

        Yields
        ------
        path : str
            The path to a ``.osu`` file.
        """
        if recurse:
            for directory, _, filenames in os.walk(path):
                for filename in filenames:
                    if filename.endswith('.osu'):
                        yield pathlib.Path(os.path.join(directory, filename))
        else:
            for entry in os.scandir(directory):
                path = entry.path
                if path.endswith('.osu'):
                    yield pathlib.Path(path)

    @classmethod
    def create_db(cls,
                  path,
                  *,
                  recurse=True,
                  cache=DEFAULT_CACHE_SIZE,
                  download_url=DEFAULT_DOWNLOAD_URL,
                  show_progress=False,
                  skip_exceptions=False):
        """Create a Library from a directory of ``.osu`` files.

        Parameters
        ----------
        path : path-like
            The path to the directory to read.
        recurse : bool, optional
            Recursively search for beatmaps?
        cache : int, optional
            The amount of beatmaps to cache in memory. This uses
            :func:`functools.lru_cache`, and if set to None will cache
            everything.
        download_url : str, optional
            The default location to download beatmaps from.
        show_progress : bool, optional
            Display a progress bar?

        Notes
        -----
        Moving the underlying ``.osu`` files invalidates the library. If this
        happens, just re-run ``create_db`` again.
        """
        path = pathlib.Path(path)
        db_path = path / '.slider.db'
        try:
            # ensure the db is cleared
            os.remove(db_path)
        except FileNotFoundError:
            pass

        self = cls(path, cache=cache, download_url=download_url)
        write_to_db = self._write_to_db

        progress = maybe_show_progress(
            self._osu_files(path, recurse=recurse),
            show_progress,
            label='Processing beatmaps: ',
            item_show_func=lambda p: 'Done!' if p is None else str(p.stem),
        )
        with self._db, progress as it:
            for path in it:
                with open(path, 'rb') as f:
                    data = f.read()

                try:
                    beatmap = Beatmap.parse(data.decode('utf-8-sig'))
                except Exception as e:
                    if skip_exceptions:
                        logging.exception(f'Failed to parse "{path}"')
                        continue
                    raise ValueError(
                        f'Failed to parse "{path}". '
                        'Use --skip-exceptions to skip this file and continue.'
                    ) from e

                write_to_db(beatmap, data, path)

        return self

    def beatmap_cached(self, *, beatmap_id=None, beatmap_md5=None):
        """Whether we have the given beatmap cached.

        Parameters
        ----------
        beatmap_id : int
            The id of the beatmap to look for.
        beatmap_md5 : str
            The md5 hash of the beatmap to look for.

        Returns
        -------
        bool
            Whether we have the given beatmap cached.
        """
        with self._db:
            if beatmap_id is not None:
                path_query = self._db.execute(
                    'SELECT 1 FROM beatmaps WHERE id = ? LIMIT 1',
                    (beatmap_id,),
                )
            else:
                path_query = self._db.execute(
                    'SELECT 1 FROM beatmaps WHERE md5 = ? LIMIT 1',
                    (beatmap_md5,),
                )

        path = path_query.fetchone()
        return bool(path)

    @staticmethod
    def _raw_read_beatmap(self, *, beatmap_id=None, beatmap_md5=None):
        """Function for opening beatmaps from disk.

        This handles both cases to only require a single lru cache.

        Notes
        -----
        This is a ``staticmethod`` to avoid a cycle from self to the lru_cache
        back to self.
        """
        with self._db:
            if beatmap_id is not None:
                key = beatmap_id
                path_query = self._db.execute(
                    'SELECT path FROM beatmaps WHERE id = ?',
                    (beatmap_id,),
                )
            else:
                key = beatmap_md5
                path_query = self._db.execute(
                    'SELECT path FROM beatmaps WHERE md5 = ?',
                    (beatmap_md5,),
                )

        path = path_query.fetchone()
        if path is None:
            raise KeyError(key)

        path, = path
        # Make path relative to the root path. We save paths relative to
        # ``self.path`` so a library can be relocated without requiring a
        # rebuild
        return Beatmap.from_path(self.path / path)

    def lookup_by_id(self, beatmap_id, *, download=False, save=False):
        """Retrieve a beatmap by its beatmap id.

        Parameters
        ----------
        beatmap_id : int or str
            The id of the beatmap to lookup.

        Returns
        -------
        beatmap : Beatmap
            The beatmap with the given id.
        download : bool. optional
            Download the map if it doesn't exist.
        save : bool, optional
            If the lookup falls back to a download, should the result be saved?

        Raises
        ------
        KeyError
            Raised when the given id is not in the library.
        """
        try:
            return self._read_beatmap(self, beatmap_id=beatmap_id)
        except KeyError:
            if not download:
                raise
            return self.download(beatmap_id, save=save)

    def lookup_by_md5(self, beatmap_md5):
        """Retrieve a beatmap by its md5 hash.

        Parameters
        ----------
        beatmap_md5 : bytes
            The md5 hash of the beatmap to lookup.

        Returns
        -------
        beatmap : Beatmap
            The beatmap with the given md5 hash.

        Raises
        ------
        KeyError
            Raised when the given md5 hash is not in the library.
        """
        return self._read_beatmap(self, beatmap_md5=beatmap_md5)

    def beatmap_from_path(self, path, copy=False):
        """Returns a beatmap from a file on disk.

        Parameters
        ----------
        path : str or pathlib.Path
            The path to the file to create the beatmap from.
        copy : bool
            Should the file be copied to the library's beatmap directory?

        Returns
        -------
        beatmap : Beatmap
            The beatmap represented by the given file.
        """
        data_bytes = open(path, 'rb').read()
        data = data_bytes.decode('utf-8-sig')
        beatmap = Beatmap.parse(data)

        if copy:
            self.save(data_bytes, beatmap=beatmap)

        return beatmap

    def save(self, data, *, beatmap=None):
        """Save raw data for a beatmap at a given location.

        Parameters
        ----------
        data : bytes
            The unparsed beatmap data.
        beatmap : Beatmap, optional
            The parsed beatmap. If not provided, the raw data will be parsed.

        Returns
        -------
        beatmap : Beatmap
            The parsed beatmap.
        """
        if beatmap is None:
            beatmap = Beatmap.parse(data.decode('utf-8-sig'))

        path = self.path / sanitize_filename(
            f'{beatmap.artist} - '
            f'{beatmap.title} '
            f'({beatmap.creator})'
            f'[{beatmap.version}]'
            f'.osu'
        )
        with open(path, 'wb') as f:
            f.write(data)

        with self._db:
            self._write_to_db(beatmap, data, path)
        return beatmap

    def delete(self, beatmap, *, remove_file=True):
        """Remove a beatmap from the library.

        Parameters
        ----------
        beatmap : Beatmap
            The beatmap to delete.
        remove_file : bool, optional
            Remove the .osu file from disk.
        """
        with self._db:
            if remove_file:
                paths = self._db.execute(
                    'SELECT path FROM beatmaps WHERE id = ?',
                    (beatmap.beatmap_id,),
                )
                for path, in paths:
                    os.unlink(path)

            self._db.execute(
                'DELETE FROM beatmaps WHERE id = ?',
                (beatmap.beatmap_id,),
            )

    def _write_to_db(self, beatmap, data, path):
        """Write data to the database.

        Parameters
        ----------
        beatmap : Beatmap
            The beatmap being stored.
        data : bytes
            The raw data for the beatmap
        path : str
            The path to save
        """
        # save paths relative to ``self.path`` so a library can be relocated
        # without requiring a rebuild
        path = path.relative_to(self.path)
        beatmap_md5 = md5(data).hexdigest()
        beatmap_id = beatmap.beatmap_id

        try:
            self._db.execute(
                'INSERT INTO beatmaps VALUES (?,?,?)',
                (beatmap_md5, beatmap_id, str(path)),
            )
        except sqlite3.IntegrityError:
            # ignore duplicate beatmaps
            pass

    def download(self, beatmap_id, *, save=False):
        """Download a beatmap.

        Parameters
        ----------
        beatmap_id : int or str
            The id of the beatmap to download.
        save : bool, optional
            Save the beatmap to disk?

        Returns
        -------
        beatmap : Beatmap
            The downloaded beatmap.
        """
        beatmap_response = requests.get(f'{self._download_url}/{beatmap_id}')
        beatmap_response.raise_for_status()

        data = beatmap_response.content
        beatmap = Beatmap.parse(data.decode('utf-8-sig'))

        if save:
            self.save(data, beatmap=beatmap)

        return beatmap

    @property
    def md5s(self):
        """All of the beatmap hashes that this has downloaded.
        """
        return tuple(
            md5 for md5, in self._db.execute('SELECT md5 FROM beatmaps')
        )

    @property
    def ids(self):
        """All of the beatmap ids that this has downloaded.
        """
        return tuple(
            int(id_)
            for id_, in self._db.execute('SELECT id FROM beatmaps')
            if id_ is not None
        )



from pathlib import Path

from slider import Beatmap


def example_beatmap(name):
    """Load one of the example beatmaps.

    Parameters
    ----------
    name : str
        The name of the example file to open.
    """
    return Beatmap.from_path(Path(__file__).parent / name)


_sendan_life_versions = frozenset({
    "Easy",
    "Normal",
    "Little's Hard",
    "Little's Insane",
    "Extra",
    "Crystal's Garakowa",
})


def sendan_life(version="Crystal's Garakowa"):
    """Load a version of the Sendan Life beatmap.

    Parameters
    ----------
    version : str
        The version to load.

    Returns
    -------
    sendan_life : Beatmap
        The beatmap object.
    """
    if version not in _sendan_life_versions:
        raise ValueError(
            f'unknown version {version}, options: {set(_sendan_life_versions)}'
        )

    return example_beatmap(
        f'Remo Prototype[CV Hanamori Yumiri] - Sendan Life (Narcissu)'
        f' [{version}].osu'
    )


_ai_no_scenario_versions = frozenset({
    "Beginner",
    "Extra",
    "Hard",
    "ktgster's Insane",
    "Kyshiro's Extra",
    "Nathan's Insane",
    "Normal",
    "pishi's Extra",
    "Sharkie's Insane",
    "sheela's Very Hard",
    "Smoothie World's Extra",
    "Super Beginner",
    "Tatoe",
    "toybot's Insane",
    "Ultra Beginner",
    "Walao's Advanced",
    "Yuistrata's Easy",
})


def miiro_vs_ai_no_scenario(version='Tatoe'):
    """Load a version of the MIIRO vs. Ai no Scenario beatmap.

    Parameters
    ----------
    version : str
        The version to load.

    Returns
    -------
    sendan_life : Beatmap
        The beatmap object.
    """
    if version not in _ai_no_scenario_versions:
        raise ValueError(
            f'unknown version {version}, options:'
            f' {set(_ai_no_scenario_versions)}'
        )

    return example_beatmap(
        f'AKINO from bless4 & CHiCO with HoneyWorks - MIIRO vs. Ai no Scenario'
        f' (monstrata) [{version}].osu'
    )



from . import Library

try:
    import click
except ImportError as e:
    raise ImportError("click must be installed to use the slider cli") from e


@click.group()
def main():
    """Slider utilities.
    """


@main.command()
@click.argument(
    'beatmaps',
    type=click.Path(exists=True, file_okay=False),
)
@click.option(
    '--recurse/--no-recurse',
    help='Recurse through ``path`` searching for beatmaps?',
    default=True,
)
@click.option(
    '--progress/--no-progress',
    help='Show a progress bar?',
    default=True,
)
@click.option(
    '--skip-exceptions/--no-skip-exceptions',
    help='Skip beatmap files that cause exceptions rather than exiting?',
    default=False,
)
def library(beatmaps, recurse, progress, skip_exceptions):
    """Create a slider database from a directory of beatmaps.
    """
    Library.create_db(
        beatmaps,
        recurse=recurse,
        show_progress=progress,
        skip_exceptions=skip_exceptions
    )


if __name__ == '__main__':
    main()



from .replay import consume_int, consume_string


class CollectionDB:
    """An osu! ``collection.db`` file.

    Parameters
    ----------
    version : int
        Version number (e.g. 20150203)
    num_collections : int
        Number of collections
    collections : list[Collection]
        List of :class:`~slider.collection.Collection` s
    """
    def __init__(self, version, num_collections, collections):
        self.version = version
        self.num_collections = num_collections
        self.collections = collections

    @classmethod
    def from_path(cls, path):
        """Read in a ``collection.db`` file from disk.

        Parameters
        ----------
        path : str or pathlib.Path
            The path to the file to read from.
        """
        with open(path, 'rb') as f:
            return cls.from_file(f)

    @classmethod
    def from_file(cls, file):
        """Read in a ``collection.db`` file from an open file object.

        Parameters
        ----------
        file : file-like
            The file object to read from.
        """
        return cls.parse(file.read())

    @classmethod
    def parse(cls, data):
        """Parse from ``collection.db`` data.

        Parameters
        ----------
        data : bytes
            The data from a ``collection.db`` file.
        """
        buffer = bytearray(data)

        version = consume_int(buffer)
        num_collections = consume_int(buffer)
        collections = []
        for i in range(num_collections):
            collections.append(Collection.parse(buffer))

        return cls(version, num_collections, collections)


class Collection:
    """An osu! collection. One or more collections are present in a
    `collection.db` file.

    Parameters
    ----------
    name : str
        The collection's name
    num_beatmaps : int
        How many beatmaps the collection contains
    md5_hashes : list[str]
        List of MD5 hashes of each beatmap
    """
    def __init__(self, name, num_beatmaps, md5_hashes):
        self.name = name
        self.num_beatmaps = num_beatmaps
        self.md5_hashes = md5_hashes

    @classmethod
    def parse(cls, buffer):
        """Parse an osu! collection.

        Parameters
        ----------
        buffer : bytearray
            Buffer passed in from parsing ``CollectionDB``
        """
        name = consume_string(buffer)
        num_beatmaps = consume_int(buffer)
        md5_hashes = []
        for i in range(num_beatmaps):
            md5_hashes.append(consume_string(buffer))

        return cls(name, num_beatmaps, md5_hashes)



import abc

# re-export
abstractmethod = abc.abstractmethod


class ABCMeta(abc.ABCMeta):
    def __new__(cls, name, bases, dict_):
        self = super().__new__(cls, name, bases, dict_)

        # copy docstrings down to implementation if the method does not
        # explicitly overwrite it
        super_dispatch = super(self, self)
        for meth in self.__abstractmethods__:
            impl = getattr(self, meth)
            if impl.__doc__ is None:
                try:
                    impl.__doc__ = getattr(super_dispatch, meth).__doc__
                except AttributeError:
                    pass

        return self



import slider.utils


def test_accuracy():
    assert slider.utils.accuracy(1, 0, 0, 0) == 1.0
    assert round(slider.utils.accuracy(0, 1, 0, 0), 4) == 0.3333
    assert round(slider.utils.accuracy(0, 0, 1, 0), 4) == 0.1667
    assert slider.utils.accuracy(0, 0, 0, 1) == 0.0
    assert round(slider.utils.accuracy(982, 100, 43, 14), 4) == 0.8977



from collections import namedtuple

from .bit_enum import BitEnum


class Mod(BitEnum):
    """The mods in osu!
    """
    no_fail = 1
    easy = 1 << 1
    no_video = 1 << 2  # not a mod anymore
    hidden = 1 << 3
    hard_rock = 1 << 4
    sudden_death = 1 << 5
    double_time = 1 << 6
    relax = 1 << 7
    half_time = 1 << 8
    nightcore = 1 << 9  # always used with double_time
    flashlight = 1 << 10
    autoplay = 1 << 11
    spun_out = 1 << 12
    relax2 = 1 << 13  # same as autopilot
    auto_pilot = 1 << 13  # same as relax2
    perfect = 1 << 14
    key4 = 1 << 15
    key5 = 1 << 16
    key6 = 1 << 17
    key7 = 1 << 18
    key8 = 1 << 19
    fade_in = 1 << 20
    random = 1 << 21
    last_mod = 1 << 22  # same as cinema
    cinema = 1 << 22  # same as last_mod
    target_practice = 1 << 23
    key9 = 1 << 24
    coop = 1 << 25
    key1 = 1 << 26
    key3 = 1 << 27
    key2 = 1 << 28
    scoreV2 = 1 << 29

    @classmethod
    def parse(cls, cs):
        """Parse a mod mask out of a list of shortened mod names.

        Parameters
        ----------
        cs : str
            The mod string.

        Returns
        -------
        mod_mask : int
            The mod mask.
        """
        if len(cs) % 2 != 0:
            raise ValueError(f'malformed mods: {cs!r}')

        cs = cs.lower()
        mapping = {
            'ez': cls.easy,
            'hr': cls.hard_rock,
            'ht': cls.half_time,
            'dt': cls.double_time,
            'hd': cls.hidden,
            'fl': cls.flashlight,
            'so': cls.spun_out,
            'nf': cls.no_fail,
        }

        mod = 0
        for n in range(0, len(cs), 2):
            try:
                mod |= mapping[cs[n:n + 2]]
            except KeyError:
                raise ValueError(f'unknown mod: {cs[n:n + 2]!r}')

        return mod


def ar_to_ms(ar):
    """Convert an approach rate value to milliseconds of time that an element
    appears on the screen before being hit.

    Parameters
    ----------
    ar : float
        The approach rate.

    Returns
    -------
    milliseconds : float
        The number of milliseconds that an element appears on the screen before
         being hit at the given approach rate.

    See Also
    --------
    :func:`slider.mod.ms_to_ar`
    """
    # NOTE: The formula for ar_to_ms is different for ar >= 5 and ar < 5
    # see: https://osu.ppy.sh/wiki/Song_Setup#Approach_Rate
    if ar >= 5:
        return 1950 - (ar * 150)
    else:
        return 1800 - (ar * 120)


def ms_to_ar(ms):
    """Convert milliseconds to hit an element into an approach rate value.

    Parameters
    ----------
    ms : float
        The number of milliseconds that an element appears on the screen before
        being hit.

    Returns
    -------
    ar : float
        The approach rate value that produces the given millisecond value.

    See Also
    --------
    :func:`slider.mod.ar_to_ms`
    """
    # NOTE: The formula for ar_to_ms is different for ar >= 5 and ar < 5
    # see: https://osu.ppy.sh/wiki/Song_Setup#Approach_Rate
    ar = (ms - 1950) / -150
    if ar < 5:
        # the ar lines cross at 5 but we use a different formula for the slower
        # approach rates.
        return (ms - 1800) / -120
    return ar


def circle_radius(cs):
    """Compute the ``CS`` attribute into a circle radius in osu! pixels.

    Parameters
    ----------
    cs : float
        The circle size.

    Returns
    -------
    radius : float
        The radius in osu! pixels.
    """
    return (512 / 16) * (1 - 0.7 * (cs - 5) / 5)


class HitWindows(namedtuple('HitWindows', 'hit_300, hit_100, hit_50')):
    """Times to hit an object at various accuracies

    Parameters
    ----------
    hit_300 : int
        The maxumium number of milliseconds away from exactly on time a hit
        can be to still be a 300
    hit_100 : int
        The maxumium number of milliseconds away from exactly on time a hit
        can be to still be a 100
    hit_50 : int
        The maxumium number of milliseconds away from exactly on time a hit
        can be to still be a 50

    Notes
    -----
    A hit further than the ``hit_50`` value away from the time of a hit object
    is a miss.
    """


def od_to_ms(od):
    """Convert an overall difficulty value into milliseconds to hit an object at
    various accuracies.

    Parameters
    ----------
    od : float
        The overall difficulty.

    Returns
    -------
    hw : HitWindows
        A namedtuple of numbers of milliseconds to hit an object at different
        accuracies.
    """
    return HitWindows(
        hit_300=(159 - 12 * od) / 2,
        hit_100=(279 - 16 * od) / 2,
        hit_50=(399 - 20 * od) / 2,
    )


def od_to_ms_300(od):
    """Convert an overall difficulty value into milliseconds to hit an object at
    maximum accuracy.

    Parameters
    ----------
    od : float
        The overall difficulty.

    Returns
    -------
    ms : float
        The number of milliseconds to hit an object at maximum accuracy.

    See Also
    --------
    :func:`slider.mod.ms_300_to_od`
    """
    return 79.5 - 6 * od


def ms_300_to_od(ms):
    """Convert the milliseconds to score a 300 into an OD value.

    Parameters
    ----------
    ms : float
        The length of the 300 window in milliseconds.

    Returns
    -------
    od : float
        The OD value that produces a 300 window of length ``ms``.

    See Also
    --------
    :func:`slider.mod.od_to_ms_300`
    """
    return (ms - 79.5) / -6



from functools import lru_cache
import datetime


class lazyval:
    """Decorator to lazily compute and cache a value.
    """
    def __init__(self, fget):
        self._fget = fget
        self._name = None

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = self._fget(instance)
        vars(instance)[self._name] = value
        return value

    def __set__(self, instance, value):
        vars(instance)[self._name] = value


class no_default:
    """Sentinel type; this should not be instantiated.

    This type is used so functions can tell the difference between no argument
    passed and an explicit value passed even if ``None`` is a valid value.

    Notes
    -----
    This is implemented as a type to make functions which use this as a default
    argument serializable.
    """
    def __new__(cls):
        raise TypeError('cannot create instances of sentinel type')


memoize = lru_cache(None)


def accuracy(count_300, count_100, count_50, count_miss):
    """Calculate osu! standard accuracy from discrete hit counts.

    Parameters
    ----------
    count_300 : int
        The number of 300's hit.
    count_100 : int
        The number of 100's hit.
    count_50 : int
        The number of 50's hit.
    count_miss : int
        The number of misses

    Returns
    -------
    accuracy : float
        The accuracy in the range [0, 1]
    """
    points_of_hits = count_300 * 300 + count_100 * 100 + count_50 * 50
    total_hits = count_300 + count_100 + count_50 + count_miss
    return points_of_hits / (total_hits * 300)


def orange(_start_or_stop, *args):
    """Range for arbitrary objects.

    Parameters
    ----------
    start, stop, step : any
        Arguments like :func:`range`.

    Yields
    ------
    value : any
        The values in the range ``[start, stop)`` with a step of ``step``.

    Notes
    -----
    ``o`` stands for object.
    """
    if not args:
        start = 0
        stop = _start_or_stop
        step = 1
    elif len(args) == 1:
        start = _start_or_stop
        stop = args[0]
        step = 1
    elif len(args) == 2:
        start = _start_or_stop
        stop, step = args
    else:
        raise TypeError(
            'orange takes from 1 to 3 positional arguments but'
            f' {len(args) + 1} were given',
        )

    while start < stop:
        yield start
        start += step


# consume_* helper functions to read osu! binary files

def consume_byte(buffer):
    result = buffer[0]
    del buffer[0]
    return result


def consume_short(buffer):
    result = int.from_bytes(buffer[:2], 'little')
    del buffer[:2]
    return result


def consume_int(buffer):
    result = int.from_bytes(buffer[:4], 'little')
    del buffer[:4]
    return result


def consume_long(buffer):
    result = int.from_bytes(buffer[:8], 'little')
    del buffer[:8]
    return result


def consume_uleb128(buffer):
    result = 0
    shift = 0
    while True:
        byte = consume_byte(buffer)
        result |= (byte & 0x7f) << shift
        if (byte & 0x80) == 0:
            break
        shift += 7

    return result


def consume_string(buffer):
    mode = consume_byte(buffer)
    if mode == 0:
        return None
    if mode != 0x0b:
        raise ValueError(
            f'unknown string start byte: {hex(mode)}, expected 0 or 0x0b',
        )
    byte_length = consume_uleb128(buffer)
    data = buffer[:byte_length]
    del buffer[:byte_length]
    return data.decode('utf-8')


_windows_epoch = datetime.datetime(1, 1, 1)


def consume_datetime(buffer):
    windows_ticks = consume_long(buffer)
    return _windows_epoch + datetime.timedelta(microseconds=windows_ticks / 10)



import slider.example_data.collections


def test_collection_simple():
    collection_db = slider.example_data.collections.test_db()
    assert collection_db.version == 20190410
    assert collection_db.num_collections == 2

    col0, col1 = collection_db.collections
    assert col0.num_beatmaps == 3
    assert col0.md5_hashes == [
        "8a67f16f3c440fa3805c14652306dfe8",
        "7b231749a908b6d17163dc3f42143774",
        "2af0646c72dee92c89649620e4cdb162",
    ]

    assert col1.num_beatmaps == 1
    assert col1.md5_hashes == ["637227d9965d5e0ba72c92a27826f5b3"]



from .beatmap import (Beatmap, Circle, Slider, Spinner, TimingPoint, HitObject,
                      HoldNote)
from .client import Client
from .game_mode import GameMode
from .mod import Mod
from .position import Position
from .replay import Replay
from .library import Library
from .collection import CollectionDB

__version__ = '0.8.2'


__all__ = [
    'Beatmap',
    'Client',
    'GameMode',
    'Library',
    'Mod',
    'Position',
    'Replay',
    'CollectionDB',
    'Circle',
    'Slider',
    'Spinner',
    'TimingPoint',
    'HitObject',
    'HoldNote'
]



import datetime
from enum import IntEnum, unique

import requests

from .game_mode import GameMode
from .mod import Mod
from .utils import lazyval, accuracy


@unique
class ApprovedState(IntEnum):
    """The state of a beatmap's approval.
    """
    loved = 4
    qualified = 3
    approved = 2
    ranked = 1
    pending = 0
    WIP = -1
    graveyard = -2


@unique
class Genre(IntEnum):
    """The genres that appear on the osu! website.
    """
    any = 0
    unspecified = 1
    video_game = 2
    anime = 3
    rock = 4
    pop = 5
    other = 6
    novelty = 7
    # note: there is no 8
    hip_hop = 9
    electronic = 10


@unique
class Language(IntEnum):
    """The languages that appear on the osu! website.
    """
    any = 0
    other = 1
    english = 2
    japanese = 3
    chinese = 4
    instrumental = 5
    korean = 6
    french = 7
    german = 8
    swedish = 9
    spanish = 10
    italian = 11


def _beatmap(self, *, save=False):
    """Lookup the associated beatmap object.

    Parameters
    ----------
    save : bool
        If the beatmap is not in the library, should it be saved?

    Returns
    -------
    beatmap : Beatmap
        The associated beatmap object.
    """
    beatmap = self._beatmap
    if beatmap is not None:
        return beatmap

    self._beatmap = beatmap = self._library.lookup_by_id(
        self.beatmap_id,
        download=True,
        save=save,
    )
    return beatmap


_beatmap.__name__ = 'beatmap'


class BeatmapResult:
    """A beatmap as represented by the osu! API.

    Parameters
    ----------
    library : Library
        The library used to store the Beatmap object.
    title : str
        The beatmap's title.
    version : str
        The beatmap's version.
    beatmap_id : int
        The beatmap_id.
    approved : ApprovedState
        The state of the beatmap's approved.
    approved_date : datetime.datetime
        The date when this map was approved.
    last_update : datetime.datetime
        The last date when this map was updated.
    star_rating : float
        The star rating for the song.
    hit_length : datetime.timedelta
        The amount of time from the first element to the last, not counting
        breaks.
    genre : Genre
        The genre that appears on the osu! website.
    language : Language
        The language that appears on the osu! website.
    total_length : datetime.timedelta
        The amount of time from the first element to the last, counting breaks.
    beatmap_md5 : str
        The md5 hash of the beatmap.
    favourite_count : int
        The number of times the beatmap has been favorited.
    play_count : int
        The number of times this beatmap has been played.
    pass_count : int
        The number of times this beatmap has been passed.
    max_combo : int
        The maximum combo that could be achieved on this beatmap.
    """
    def __init__(self,
                 library,
                 title,
                 version,
                 beatmap_id,
                 approved,
                 approved_date,
                 last_update,
                 star_rating,
                 hit_length,
                 genre,
                 language,
                 total_length,
                 beatmap_md5,
                 favourite_count,
                 play_count,
                 pass_count,
                 max_combo):
        self._library = library
        self.title = title
        self.version = version
        self.beatmap_id = beatmap_id
        self.approved = approved
        self.approved_date = approved_date
        self.last_update = last_update
        self.star_rating = star_rating
        self.hit_length = hit_length
        self.genre = genre
        self.language = language
        self.total_length = total_length
        self.beatmap_md5 = beatmap_md5
        self.favourite_count = favourite_count
        self.play_count = play_count
        self.pass_count = pass_count
        self.max_combo = max_combo

        self._beatmap = None

    beatmap = _beatmap

    def __repr__(self):
        return f'<{type(self).__qualname__}: {self.title} [{self.version}]>'


class UserEvent:
    """Recent events for a user.

    Parameters
    -----------
    library : Library
        The library used to store the Beatmap object.
    display_html : str
        The html to display on the osu! site.
    beatmap_id : int
        The beatmap_id of the event.
    beatmapset_id : int
        The beatmapset_id of the event.
    date : datetime.date
        The date of the event.
    epic_factor : int
        How epic was this event.
    """
    def __init__(self,
                 library,
                 display_html,
                 beatmap_id,
                 beatmapset_id,
                 date,
                 epic_factor):
        self._library = library
        self.display_html = display_html
        self.beatmap_id = beatmap_id
        self.beatmapset_id = beatmapset_id
        self.date = date
        self.epic_factor = epic_factor

        self._beatmap = None

    beatmap = _beatmap


class User:
    """Information about an osu! user.

    Parameters
    ----------
    client : Client
        The client needed to make further requests.
    user_id : int
        The user id.
    user_name : str
        The user name.
    count_300 : int
        The total number of 300s ever hit.
    count_100 : int
        The total number of 100s ever hit.
    count_50 : int
        The total number of 50s ever hit.
    play_count : int
        The total number of plays.
    ranked_score : int
        The user's ranked score.
    total_score : int
        The user's total score.
    pp_rank : int
        The user's rank with the PP system.
    level : float
        The user's level.
    pp_raw : float
        The user's total unweighted PP.
    accuracy : float
        The user's ranked accuracy.
    count_ss : int
        The number of SSs scored.
    count_s : int
        The number of Ss scored.
    count_a : int
        The number of As scored.
    country : str
        The country code for the user's home country.
    pp_country_rank : int
        The user's rank with the PP system limited to other players in their
        country.
    events : list[UserEvent]
        Recent user events.
    game_mode : GameMode
        The game mode the user information is for.
    """
    def __init__(self,
                 client,
                 user_id,
                 user_name,
                 count_300,
                 count_100,
                 count_50,
                 play_count,
                 ranked_score,
                 total_score,
                 pp_rank,
                 level,
                 pp_raw,
                 accuracy,
                 count_ss,
                 count_s,
                 count_a,
                 country,
                 pp_country_rank,
                 events,
                 game_mode):
        self._client = client
        self.user_id = user_id
        self.user_name = user_name
        self.count_300 = count_300
        self.count_100 = count_100
        self.count_50 = count_50
        self.play_count = play_count
        self.ranked_score = ranked_score
        self.total_score = total_score
        self.pp_rank = pp_rank
        self.level = level
        self.pp_raw = pp_raw
        self.accuracy = accuracy
        self.count_ss = count_ss
        self.count_s = count_s
        self.count_a = count_a
        self.country = country
        self.pp_country_rank = pp_country_rank
        self.events = events
        self.game_mode = game_mode

    def __repr__(self):
        return (
            f'<{type(self).__qualname__}: {self.user_name}'
            f' ({self.game_mode.name})>'
        )

    def high_scores(self, limit=10):
        """Lookup the user's high scores.

        Parameters
        ----------
        limit : int, optional
            The number of scores to look up.

        Returns
        -------
        high_scores : list[HighScore]
            The user's high scores.
        """
        return self._client.user_best(
            user_id=self.user_id,
            limit=limit,
            game_mode=self.game_mode,
            _user_ob=self,
        )


class HighScore:
    """A high score for a user or beatmap.

    Parameters
    ----------
    client : Client
        The client used to make further requests.
    beatmap_id : int
        The beatmap_id of the map this is a score for.
    score : int
        The score earned in this high score.
    max_combo : int
        The max combo.
    count_300 : int
        The number of 300s in the high score.
    count_100 : int
        The number of 100s in the high score.
    count_50 : int
        The number of 50s in the high score.
    count_miss : int
        The number of misses in the high score.
    count_katu : int
        The number of katu in the high score.
    count_geki : int
        The number of geki in the high score.
    perfect : bool
        Did the user fc the map?
    mods : set[Mod]
        The mods used.
    user_id : int
        The id of the user who earned this high score.
    rank : str
        The letter rank earned. A suffix ``H`` means hidden or flashlight was
        used, like a silver S(S).
    pp : float
        The unweighted PP earned for this high score.
    """
    def __init__(self,
                 client,
                 beatmap_id,
                 score,
                 max_combo,
                 count_300,
                 count_100,
                 count_50,
                 count_miss,
                 count_katu,
                 count_geki,
                 perfect,
                 mods,
                 user_id,
                 date,
                 rank,
                 pp,
                 _user=None):

        self._client = client
        self._library = client.library
        if _user is not None:
            self.user = _user

        self.beatmap_id = beatmap_id
        self.score = score
        self.max_combo = max_combo
        self.count_300 = count_300
        self.count_100 = count_100
        self.count_50 = count_50
        self.count_miss = count_miss
        self.count_katu = count_katu
        self.count_geki = count_geki
        self.perfect = perfect
        self.mods = mods
        self.user_id = user_id
        self.date = date
        self.rank = rank
        self.pp = pp

        self._beatmap = None

    beatmap = _beatmap

    @lazyval
    def user(self):
        return self._client.user(user_id=self.user_id)

    @property
    def accuracy(self):
        return accuracy(
            self.count_300,
            self.count_100,
            self.count_50,
            self.count_miss,
        )

    def __repr__(self):
        return (
            f'<{type(self).__qualname__} user_id={self.user_id};'
            f' beatmap_id=self.beatmap_id>'
        )


class UnknownBeatmap(LookupError):
    """Raised when a beatmap id or md5 is not known.

    Parameters
    ----------
    kind : {'id', 'md5'}
        The kind of identifier used.
    id_ : str
        The unknown beatmap id.
    """
    def __init__(self, kind, id_):
        self.kind = kind
        self.id_ = id_

    def __str__(self):
        return f'no beatmap found that matched {self.kind}: {self.id_}'


class Client:
    """A client for interacting with the osu! rest API.

    Parameters
    ----------
    library : Library
        The library used to look up or cache beatmap objects.
    api_key : str
        The api key to use.
    """
    DEFAULT_API_URL = 'https://osu.ppy.sh/api'

    def __init__(self, library, api_key, api_url=DEFAULT_API_URL):
        self.library = library
        self.api_key = api_key
        self.api_url = api_url

    def copy(self):
        """Create a copy suitable for use in a new thread.

        Returns
        -------
        Client
            The new copy.
        """
        return type(self)(
            library=self.library.copy(),
            api_key=self.api_key,
            api_url=self.api_url,
        )

    @staticmethod
    def _user_and_type(user_name, user_id, *, required):
        """Normalize user_name or user_id into the 'u' and 't' fields.

        Parameters
        ----------
        user_name : str or None
            The user name.
        user_id : int or None
            The user id.
        required : bool
            Is user information required?

        Returns
        -------
        user_info : (any, str) or None
            The user identifier and type or None if not provided and required
            is False.
        """
        if user_name is not None and user_id is not None:
            raise ValueError('only one of user_name or user_id can be passed')

        if user_name is not None:
            user = user_name
            type_ = 'string'
        elif user_id is not None:
            user = user_id
            type_ = 'id'
        elif required:
            raise ValueError('one of user_name or user_id must be passed')
        else:
            return None

        return user, type_

    # differences in the osu! api names and slider's names
    _beatmap_aliases = {
        'beatmapset_id': 'beatmap_set_id',
        'difficultyrating': 'star_rating',
        'diff_size': 'circle_size',
        'diff_overall': 'overall_difficulty',
        'diff_approach': 'approach_rate',
        'diff_drain': 'health_drain',
        'genre_id': 'genre',
        'language_id': 'language',
        'file_md5': 'beatmap_md5',
        'playcount': 'play_count',
        'passcount': 'pass_count',
    }

    def _parse_date(cs):
        return datetime.datetime.strptime(cs, '%Y-%m-%f %H:%M:%S')

    def _parse_timedelta(cs):
        return datetime.timedelta(seconds=int(cs))

    def _parse_optional(class_):
        def func(cs):
            if cs is None:
                return cs
            return class_(cs)
        return func

    def _identity(cs):
        return cs

    _beatmap_conversions = {
        'approved': lambda cs: ApprovedState(int(cs)),
        'approved_date': _parse_date,
        'last_update': _parse_date,
        'beatmap_id': int,
        'star_rating': float,
        'hit_length': _parse_timedelta,
        'genre': lambda cs: Genre(int(cs)),
        'language': lambda cs: Language(int(cs)),
        'total_length': _parse_timedelta,
        'beatmap_md5': _identity,
        'favourite_count': int,
        'play_count': int,
        'pass_count': int,
        'max_combo': _parse_optional(int),
        'title': _identity,
        'version': _identity,
    }

    def beatmap(self,
                *,
                since=None,
                beatmap_set_id=None,
                beatmap_id=None,
                beatmap_md5=None,
                user_id=None,
                user_name=None,
                game_mode=None,
                include_converted_beatmaps=False,
                limit=500):
        """Retrieve information about a beatmap or set of beatmaps from the
        osu! API.

        Parameters
        ----------
        since : datetime.date, optional
            Return all beatmaps ranked since this date.
        beatmap_set_id : str, optional
            Return all beatmaps in this set.
        beatmap_id : str, optional
            Return the single beatmap with the given id.
        beatmap_md5 : str, optional
            Return the single beatmap with the given hash.
        user_id : int, optional
            Return beatmaps made by the given user.
        user_name : str, optional
            Return beatmaps made by the given user.
        game_mode : GameMode, optional
            Return beatmaps for the given game mode.
        include_converted_beatmaps : bool, optional
            Return converted beatmaps. This only applies for non-osu!standard
            game modes.
        limit : int, optional
            The maximum number of beatmaps to return. This must be less than
            500.
        """
        beatmap_identifiers = {
            k: v for k, v in {
                'beatmap_set_id': beatmap_set_id,
                'beatmap_id': beatmap_id,
                'beatmap_md5': beatmap_md5,
            }.items()
            if v is not None
        }
        if len(beatmap_identifiers) > 1:
            raise ValueError(
                f'only one of beatmap_set_id, beatmap_id, or beatmap_md5 can'
                f' be passed, got {beatmap_identifiers!r}',
            )

        if limit > 500:
            raise ValueError('only 500 beatmaps can be requested at one time')

        parameters = {
            'k': self.api_key,
            'a': int(bool(include_converted_beatmaps)),
            'limit': limit,
        }

        if since is not None:
            parameters['since'] = since.isoformat()

        if beatmap_set_id is not None:
            parameters['s'] = beatmap_set_id
        elif beatmap_id is not None:
            parameters['b'] = beatmap_id
        elif beatmap_md5 is not None:
            parameters['h'] = beatmap_md5

        user_info = self._user_and_type(user_name, user_id, required=False)
        if user_info is not None:
            parameters['u'], parameters['t'] = user_info

        if game_mode is not None:
            parameters['m'] = int(game_mode)

        response = requests.get(
            f'{self.api_url}/get_beatmaps',
            params=parameters,
        )
        response.raise_for_status()

        as_dicts = (
            {self._beatmap_aliases.get(k, k): v for k, v in beatmap.items()}
            for beatmap in response.json()
        )
        converted = [
            BeatmapResult(
                library=self.library,
                **{
                    k: self._beatmap_conversions[k](v)
                    for k, v in d.items()
                    if k in self._beatmap_conversions
                },
            )
            for d in as_dicts
        ]

        if beatmap_id is not None or beatmap_md5 is not None:
            try:
                converted, = converted
            except ValueError:
                if beatmap_id is not None:
                    kind = 'id'
                    id_ = beatmap_id
                else:
                    kind = 'md5'
                    id_ = beatmap_md5

                raise UnknownBeatmap(kind, id_)

        return converted

    _user_aliases = {
        'username': 'user_name',
        'count300': 'count_300',
        'count100': 'count_100',
        'count50': 'count_50',
        'playcount': 'play_count',
        'count_rank_ss': 'count_ss',
        'count_rank_s': 'count_s',
        'count_rank_a': 'count_a',
    }

    def _parse_user_events(events, _parse_date=_parse_date):
        out = []
        for event in events:
            # beatmap id can be null if the event is a supporter gift
            if event['beatmap_id'] is not None:
                event['beatmap_id'] = int(event['beatmap_id'])
            event['date'] = _parse_date(event['date'])
            event['epic_factor'] = event.pop('epicfactor')
            out.append(event)

        return out

    _user_conversions = {
        'user_id': int,
        'user_name': _identity,
        # these attributes can be null for users that have never played before,
        # see user 17906393 / #97 on github
        'count_300': _parse_optional(int),
        'count_100': _parse_optional(int),
        'count_50': _parse_optional(int),
        'play_count': _parse_optional(int),
        'ranked_score': _parse_optional(int),
        'total_score': _parse_optional(int),
        'pp_rank': _parse_optional(int),
        'level': _parse_optional(float),
        'pp_raw': _parse_optional(float),
        'accuracy': _parse_optional(float),
        'count_ss': _parse_optional(int),
        'count_s': _parse_optional(int),
        'count_a': _parse_optional(int),
        'country': _identity,
        'pp_country_rank': _parse_optional(int),
        'events': _parse_user_events,
    }

    def user(self,
             *,
             user_name=None,
             user_id=None,
             game_mode=GameMode.standard,
             event_days=1):
        """Retrieve information about a user.

        Parameters
        ----------
        user_name : str
            The name of the user to look up. This cannot be passed with
            ``user_id``.
        user_id : int
            The id of the user to look up. This cannot be passed with
            ``user_name``
        game_mode : GameMode, optional
            The game mode to look up stats for.
        event_days : int, optional
            Max number of days between now and last event date in the range
            [1, 31].

        Returns
        -------
        user : User
            The requested user.
        """
        if user_name is not None and user_id is not None:
            raise ValueError('only one of user_name or user_id can be passed')

        user, type_ = self._user_and_type(user_name, user_id, required=True)

        if not (1 <= event_days <= 31):
            raise ValueError(
                f'event_days must be in range [1, 31], got {event_days!r}',
            )

        response = requests.get(
            self.api_url + '/get_user',
            params={
                'k': self.api_key,
                'u': user,
                't': type_,
                'm': int(game_mode),
                'event_days': event_days,
            },
        )
        response.raise_for_status()

        dict_, = (
            {self._user_aliases.get(k, k): v for k, v in beatmap.items()}
            for beatmap in response.json()
        )

        library = self.library
        events = dict_['events']

        for event in events:
            event['library'] = library

        return User(
            client=self,
            **{
                k: self._user_conversions[k](v)
                for k, v in dict_.items()
                if k in self._user_conversions
            },
            game_mode=game_mode,
        )
        return user

    _user_best_aliases = {
        'username': 'user_name',
        'maxcombo': 'max_combo',
        'count300': 'count_300',
        'count100': 'count_100',
        'count50': 'count_50',
        'countmiss': 'count_miss',
        'countkatu': 'count_katu',
        'countgeki': 'count_geki',
        'enabled_mods': 'mods',
    }

    _user_best_conversions = {
        'beatmap_id': int,
        'score': int,
        'user_name': _identity,
        'max_combo': int,
        'count_300': int,
        'count_100': int,
        'count_50': int,
        'count_miss': int,
        'count_katu': int,
        'count_geki': int,
        'perfect': lambda m: bool(int(m)),
        'mods': lambda m: {v for v in Mod.unpack(int(m)).values() if v},
        'user_id': int,
        'date': _parse_date,
        'rank': _identity,
        'pp': float,
    }

    def user_best(self,
                  *,
                  user_name=None,
                  user_id=None,
                  game_mode=GameMode.standard,
                  limit=10,
                  _user_ob=None):
        """Retrieve information about a user's best scores.

        Parameters
        ----------
        user_name : str
            The name of the user to look up. This cannot be passed with
            ``user_id``.
        user_id : int
            The id of the user to look up. This cannot be passed with
            ``user_name``
        game_mode : GameMode, optional
            The game mode to look up stats for. Defaults to osu! standard.
        limit : int, optional
            The number of results to return in the range [1, 100].

        Returns
        -------
        high_scores : list[HighScore]
            The user's best scores.
        """
        user, type_ = self._user_and_type(user_name, user_id, required=True)

        if not (1 <= limit <= 100):
            raise ValueError(
                f'limit must be in the range [1, 100], got: {limit!r}',
            )

        response = requests.get(
            self.api_url + '/get_user_best',
            params={
                'k': self.api_key,
                'u': user,
                't': type_,
                'm': int(game_mode),
                'limit': limit,
            },
        )
        response.raise_for_status()

        as_dicts = (
            {self._user_best_aliases.get(k, k): v for k, v in beatmap.items()}
            for beatmap in response.json()
        )
        return [
            HighScore(
                client=self,
                **{
                    k: self._user_best_conversions[k](v)
                    for k, v in d.items()
                    if k in self._user_best_conversions
                },
                _user=_user_ob,
            )
            for d in as_dicts
        ]



import bisect
import math
from itertools import accumulate, chain

import numpy as np
try:  # SciPy >= 0.19
    from scipy.special import comb
except ImportError:
    from scipy.misc import comb

from .abc import ABCMeta, abstractmethod
from .position import Position
from .utils import lazyval


class Curve(metaclass=ABCMeta):
    """Curves are functions that define the path that a slider takes from start
    to stop.

    Parameters
    ----------
    points : list[Position]
        The positions of the control points.
    req_length : float
        The pixel length of the curve.
    """
    def __init__(self, points, req_length):
        self.points = points
        self.req_length = req_length

    @classmethod
    def from_kind_and_points(cls, kind, points, req_length):
        if kind == 'B':
            return MultiBezier(points, req_length)
        if kind == 'L':
            return Linear(points, req_length)
        if kind == 'C':
            return Catmull(points, req_length)
        if kind == 'P':
            if len(points) != 3:
                # it seems osu! uses the bezier curve if there are more than 3
                # points
                # https://github.com/ppy/osu/blob/7fbbe74b65e7e399072c198604e9db09fb729626/osu.Game/Rulesets/Objects/SliderCurve.cs#L32  # noqa
                return MultiBezier(points, req_length)

            try:
                center = get_center(*points)
            except ValueError:
                # we cannot use a perfect curve function for collinear points;
                # osu! also falls back to a bezier here
                return MultiBezier(points, req_length)

            return Perfect(points, req_length, _center=center)

        raise ValueError(f'unknown curve kind: {kind!r}')

    @abstractmethod
    def __call__(self, t):
        """Compute the position of the curve at time ``t``.

        Parameters
        ----------
        t : float
            The time along the distance of the curve in the range [0, 1]

        Returns
        -------
        position : Position
            The position of the curve.
        """
        raise NotImplementedError('__call__')

    def pack(self):
        """The packed string representing this curve in ``.osu`` file,
        which is a part of the packed string of a ``beatmap.Slider``
        hit object.

        Returns
        -------
        packed_str : str
            The packed str of this curve.
        """
        if isinstance(self, Catmull):
            kind = 'C'
        elif isinstance(self, Linear):
            kind = 'L'
        elif isinstance(self, Perfect):
            kind = 'P'
        else:
            # Bezier
            kind = 'B'
        # The first point is specified at the beginning of
        # HitObject's packed str and should not be included here
        return '|'.join(chain([kind],
                              (str(int(point.x)) + ':' + str(int(point.y))
                               for point in self.points[1:])))

    @lazyval
    def hard_rock(self):
        """This curve when played with hard rock.
        """
        return type(self)(
            [Position(p.x, 384 - p.y) for p in self.points],
            self.req_length,
        )


class Bezier(Curve):
    """Bezier curves of degree ``len(points)``.

    Parameters
    ----------
    points : list[Position]
        The positions of the control points.
    req_length : float
        The pixel length of the curve.
    """
    def __init__(self, points, req_length):
        super().__init__(points, req_length)
        self._coordinates = np.array(points).T

    def __call__(self, t):
        coordinates = self.at(t * (self.req_length / self.length))
        return Position(coordinates[0, 0], coordinates[0, 1])

    def at(self, t):
        """Compute the positions of the bezier curve for a sequence of times
        from 0 to 1.

        Parameters
        ----------
        t : np.ndarray
            The times to compute for.

        Returns
        -------
        coordinates : np.ndarray
            A two dimensional array of shape (len(t), 2) where the first column
            holds the x coordinate at time t and the second column holds the y
            coordinate at time t.
        """
        t = np.asarray(t).reshape(-1, 1)

        points = self.points

        n = len(points) - 1
        ixs = np.arange(n + 1)
        return np.sum(
            (
                comb(n, ixs) *
                (1 - t) ** (n - ixs) *
                t ** ixs
            )[:, np.newaxis] *
            self._coordinates,
            axis=-1,
        )

    @lazyval
    def length(self):
        """Approximates length as piecewise linear function.
        """
        # NOTE: if the error is high, try increasing the samples
        points = self.at(np.linspace(0, 1, 50))
        return np.sum(
            np.sqrt(
                np.sum(
                    np.square(
                        np.diff(points, axis=0),
                    ),
                    axis=1,
                ),
            ),
        )


class _MetaCurveMixin:
    """Mixin for defining curves which are collections of other curves.

    Implementers must provide a ``_curves`` attribute which is the collection
    of curves.
    """
    @lazyval
    def _ts(self):
        lengths = [c.length for c in self._curves]
        length = sum(lengths)
        out = []
        for i, j in enumerate(accumulate(lengths[:-1])):
            self._curves[i].req_length = lengths[i]
            out.append(j / length)
        self._curves[-1].req_length = max(
            0,
            lengths[-1] - (length - self.req_length),
        )
        out.append(1)
        return out

    def __call__(self, t):
        ts = self._ts
        if len(self._curves) == 1:
            # Special case where we only have one curve
            return self._curves[0](t)

        bi = bisect.bisect_left(ts, t)
        if bi == 0:
            pre_t = 0
        else:
            pre_t = ts[bi - 1]

        post_t = ts[bi]

        return self._curves[bi]((t - pre_t) / (post_t - pre_t))


class MultiBezier(_MetaCurveMixin, Curve):
    """MultiBezier is a collection of `:class:~slider.curve.Bezier` curves
    stitched together at the ends.

    Parameters
    ----------
    points : list[Position]
        The positions of the control points.
    req_length : float
        The pixel length of the curve.
    """
    def __init__(self, points, req_length):
        super().__init__(points, req_length)

        self._curves = [
            Bezier(subpoints, None)
            for subpoints in self._split_at_dupes(points)
        ]

    @staticmethod
    def _split_at_dupes(input_):
        """Split a sequence of points on duplicates.

        Parameters
        ----------
        inp : list[any]
            The input sequence to split.

        Yields
        ------
        group : list[any]
            The groups split on duplicates.
        """
        old_ix = 0
        for n, (a, b) in enumerate(zip(input_, input_[1:]), 1):
            if a == b:
                yield input_[old_ix:n]
                old_ix = n

        tail = input_[old_ix:]
        if tail:
            yield tail


class Linear(_MetaCurveMixin, Curve):
    """Linear curves traverse the line segments between the given points.

    Parameters
    ----------
    points : list[Position]
        The positions of the control points.
    req_length : float
        The pixel length of the curve.
    """
    def __init__(self, points, req_length):
        super().__init__(points, req_length)

        self._curves = [
            Bezier(subpoints, None) for subpoints in zip(points, points[1:])
        ]


class Perfect(Curve):
    """Perfect curves traverse the circumference of the circle that passes
    through the given points.

    Parameters
    ----------
    points : list[Position]
        The positions of the control points.
    req_length : float
        The pixel length of the curve.
    """
    def __init__(self, points, req_length, *, _center=None):
        # NOTE: _center exists as an optimization to not recompute the center
        # of the circle when calling ``Curve.from_kind_and_points``.
        self.points = points
        self.req_length = req_length

        if _center is None:
            _center = get_center(*points)

        self._center = _center

        coordinates = np.array(points) - _center

        # angles of 3 points to center
        start_angle, end_angle = np.arctan2(
            coordinates[::2, 1],
            coordinates[::2, 0],
        )

        # normalize so that self._angle is positive
        if end_angle < start_angle:
            end_angle += 2 * math.pi

        # angle of arc sector that describes slider
        self._angle = end_angle - start_angle

        # switch angle direction if necessary
        a_to_c = coordinates[2] - coordinates[0]
        ortho_a_to_c = np.array((a_to_c[1], -a_to_c[0]))
        if np.dot(ortho_a_to_c, coordinates[1] - coordinates[0]) < 0:
            self._angle = -(2 * math.pi - self._angle)

        length = abs(
            self._angle *
            math.sqrt(coordinates[0][0] ** 2 + coordinates[0][1] ** 2),
        )
        if length > req_length:
            self._angle *= req_length / length

    def __call__(self, t):
        return rotate(self.points[0], self._center, self._angle * t)


class Catmull(Curve):
    """Catmull curves implement the Catmull-Rom spline algorithm for defining
    the path.
    """
    def __init__(self, points, req_length):
        super().__init__(points, req_length)
        points = np.array(points)

        # implementation follows notes at https://cubic.org/docs/hermite.htm
        self.h = np.array([[2,  -2,  1,  1],
                           [-3,  3, -2, -1],
                           [0,   0,  1,  0],
                           [1,   0,  0,  0]])

        tangents_x = []
        tangents_y = []
        self.Cxs = []
        self.Cys = []

        # A catmull slider with a single control point doesn't make a whole lot
        # of sense to me.. We'll treat it as a slider with 0 duration and 0
        # length; ie, its position is always the position of the single control
        # point.
        #
        # This case has only ever come up in aspire maps. The goal here is
        # primarily to avoid raising an index error, not being 100% correct.
        #
        # lazer might do something more intelligent here. Its implementation is
        # at https://github.com/ppy/osu-framework/blob/4de1bae5d0b0a7fb95009498
        # fd4a960715be2419/osu.Framework/Utils/PathApproximator.cs#L142 and may
        # be worth checking against in the future.
        # see also https://github.com/llllllllll/slider/issues/108.
        if len(points) == 1:
            return

        # The tangent for point i is defined as 0.5 * (P_(i + 1) - P_(i - 1)),
        # so we need to consider the point behind it and the point in front of
        # it. We:
        # * roll points right by one. Then we replace first element with
        #   the previous first element (which is now second) so the first
        #   tangent can be calculated properly.
        # * roll points left by one. Then we replace last element with the
        #   previous last element (which is now second to last) so the last
        #   tangent can be calculated properly.
        # to create the p_aheads and p_behinds lists respectively.
        # For example, in a list of five points, we have:
        # * points =    [p1, p2, p3, p4, p5]
        # * p_aheads =  [p2, p3, p4, p5, p5]
        # * p_behinds = [p1, p1, p2, p3, p4]

        p_aheads = np.roll(points, 1)
        p_aheads[0] = p_aheads[1]

        p_behinds = np.roll(points, -1)
        p_behinds[-1] = p_behinds[-2]

        # we interpolate x and y separately, so track their tangents in two
        # separate lists.
        for (p_ahead, p_behind) in zip(p_aheads, p_behinds):
            tangent = 0.5 * (p_ahead[0] - p_behind[0])
            tangents_x.append(tangent)

            tangent = 0.5 * (p_ahead[1] - p_behind[1])
            tangents_y.append(tangent)

        # for each curve we consider its start and end point (and start and end
        # tangent). This means the number of curves will be one less than the
        # number of points.
        for ((p1, p2), (t1, t2)) in zip(zip(points, points[1:]),
                                        zip(tangents_x, tangents_x[1:])):
            Cx = np.array([p1[0], p2[0], t1, t2])
            # make it a column vector
            Cx = Cx[:, np.newaxis]
            self.Cxs.append(Cx)

        for ((p1, p2), (t1, t2)) in zip(zip(points, points[1:]),
                                        zip(tangents_y, tangents_y[1:])):
            Cy = np.array([p1[1], p2[1], t1, t2])
            # make it a column vector
            Cy = Cy[:, np.newaxis]
            self.Cys.append(Cy)

    def __call__(self, t):
        # single control point case. Treat as a 0 length, 0 duration slider,
        # with position equal to its single control point. See comment in
        # __init__ for details.
        if len(self.points) == 1:
            return self.points[0]

        # for consistency with website notes linked above
        s = t
        S = np.array([s ** 3, s ** 2, s, 1])
        # catmull curves are made up of a number of individual curves. Assuming
        # osu! weights each curve equally (that is, each curve takes an equal
        # amount of time to traverse regardless of its size), we can get the
        # curve that should be used for a certain t by multiplying by the
        # number of curves and rounding up.
        curve_index = math.ceil(t * len(self.Cxs)) - 1
        Cx = self.Cxs[curve_index]
        Cy = self.Cys[curve_index]

        px = (S @ self.h) @ Cx
        py = (S @ self.h) @ Cy
        # A bit of dimensional analysis:
        # S = 1x4
        # C = 4x1
        # h = 4x4
        #
        # P = (S * h) * C = (1x4 * 4x4) * 4x1 = 1x4 * 4x1 = 1x1
        # Result of multiplication is a 1x1 ndarray so convert to a float.
        px = float(px)
        py = float(py)
        return Position(px, py)


def get_center(a, b, c):
    """Returns the Position of the center of the circle described by the 3
    points

    Parameters
    ----------
    a, b, c : Position
        The three positions.

    Returns
    -------
    center : Position
        The center of the three points.

    Notes
    -----
    This uses the same algorithm as osu!
    https://github.com/ppy/osu/blob/7fbbe74b65e7e399072c198604e9db09fb729626/osu.Game/Rulesets/Objects/CircularArcApproximator.cs#L23  # noqa
    """
    a, b, c = np.array([a, b, c], dtype=np.float64)

    a_squared = np.sum(np.square(b - c))
    b_squared = np.sum(np.square(a - c))
    c_squared = np.sum(np.square(a - b))

    if np.isclose([a_squared, b_squared, c_squared], 0).any():
        raise ValueError(f'given points are collinear: {a}, {b}, {c}')

    s = a_squared * (b_squared + c_squared - a_squared)
    t = b_squared * (a_squared + c_squared - b_squared)
    u = c_squared * (a_squared + b_squared - c_squared)

    sum_ = s + t + u

    if np.isclose(sum_, 0):
        raise ValueError(f'given points are collinear: {a}, {b}, {c}')

    return Position(*(s * a + t * b + u * c) / sum_)


def rotate(position, center, radians):
    """Returns a Position rotated r radians around centre c from p
    Parameters
    ----------
    position : Position
        The position to rotate.
    center : Position
        The point to rotate about.
    radians : float
        The number of radians to rotate ``position`` by.
    """
    p_x, p_y = position
    c_x, c_y = center

    x_dist = p_x - c_x
    y_dist = p_y - c_y

    return Position(
        (x_dist * math.cos(radians) - y_dist * math.sin(radians)) + c_x,
        (x_dist * math.sin(radians) + y_dist * math.cos(radians)) + c_y,
    )



from collections import namedtuple
import numpy as np


class Position(namedtuple('Position', 'x y')):
    """A position on the osu! screen.

    Parameters
    ----------
    x : int or float
        The x coordinate in the range.
    y : int or float
        The y coordinate in the range.

    Notes
    -----
    The visible region of the osu! standard playfield is [0, 512] by [0, 384].
    Positions may fall outside of this range for slider curve control points.
    """
    x_max = 512
    y_max = 384

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


class Point(namedtuple('Point', 'x y offset')):
    """A position and time on the osu! screen.

    Parameters
    ----------
    x : int or float
        The x coordinate in the range.
    y : int or float
        The y coordinate in the range.
    offset : timedelta
        The time

    Notes
    -----
    The visible region of the osu! standard playfield is [0, 512] by [0, 384].
    Positions may fall outside of this range for slider curve control points.
    """


def distance(start, end):
    return np.sqrt((start.x - end.x) ** 2 + (start.y - end.y) ** 2)
